import sqlalchemy as db
from dp_sqlalchemy_wrapper import makeBase
from sqlalchemy.orm import relationship

"""
LogicalEntity <--> LogicalEntityTranlsationInTargetLanguage <--> TargetLanguage 

"""
TranslationBase = makeBase()

class LogicalEntity(TranslationBase ):
    logical_key = db.Column(db.String, unique=True)
    translation_relations = relationship("LogicalEntityLanguageTranslationRelation", back_populates="logical_entity")

class LogicalEntityLanguageTranslationRelation(TranslationBase ):
    logical_key_ref = db.Column(db.String, nullable=False)
    code_ref = db.Column(db.String, nullable=False)
    translation = db.Column(db.String, nullable=False)
    language = relationship("Language", back_populates="translation_relations")
    logical_entity = relationship("LogicalEntity", back_populates="translation_relations")

class Language(TranslationBase):
    code = db.Column(db.String(2), unique=True, nullable=False) # ISO-639-1-Code
    translation_relations = relationship("LogicalEntityLanguageTranslationRelation", back_populates="language")



from .key_logic import TranslationKeyBuilder, TranslationKeyRetriever
from .model_mixin import TranslatableModelMixin
from .models import TranslationBase, LogicalEntity, LogicalEntityLanguageTranslationRelation, Language

__all__ = [
    TranslationKeyRetriever, TranslationKeyBuilder, TranslatableModelMixin,
    TranslationBase, LogicalEntity, LogicalEntityLanguageTranslationRelation, Language
]


import re
from sqlalchemy.orm import Session
from typing import List

TranslationKey_SEPERATOR =  "_%_"

class TranslationKeyBuilder:
    NULL_VALUE = "null"
    SEPERATOR = "_%_"
    REGEX = re.compile(f"^\\w+({TranslationKey_SEPERATOR}\\w+){{3}}$")
    KW_NAMES = ["context", "concept", "instance_id", "attr"]
    

    @classmethod
    def construct_translation_key(cls,
        context: str=None,
        concept: str=None,
        instance_id: str=None,
        attr: str=None
        ):
        """Construct a translation key

        Args:
            context (str): The context in which the translation is required, e.g. a specific database
            concept (str): The concept for which the translation is required, e.g. a specific table
            instance_id (str): A key that uniquely identifies an instance of the specific concept
            attr (str): The name of the attribute that should be translated for the specific instance

        Returns:
            [str]: A translation_key that is unique within the whole translation database and can be used to find
                   translations for several target_languages
        """
        return cls.SEPERATOR.join([context, concept, instance_id, attr])
    
    @classmethod
    def destruct_translation_key(cls, tranlsation_key: str):
        match_result_str = cls.REGEX.match(tranlsation_key).group(0)
        assert  match_result_str == tranlsation_key
        return dict(zip(cls.KW_NAMES, match_result_str.split(cls.SEPERATOR)))


class TranslationKeyRetriever:

    @classmethod
    def get_all_translation_keys_of_translateable_model(cls, session: Session, model: type) -> List[str]:
        translation_keys = []
        translation_keys.append(model.get_logical_key_for_concept())

        for attr in model.get_translateable_attrs():
            translation_keys.append(model.get_logical_key_for_attr(attr))

        for instance in session.query(model).all():
            for attr in model.get_translateable_attrs():
                translation_keys.append(instance.get_logical_key_for_attr_value(attr))
        return translation_keys


from .key_logic import TranslationKeyBuilder
from typing import List
from abc import abstractmethod

class TranslatableModelMixin:

    def get_logical_key_for_attr_value(self, attr_name: str):
        return TranslationKeyBuilder.construct_translation_key(
            context=self.get_translation_context(),
            concept=type(self).__name__,
            instance_id=self.choose_instance_id(),
            attr=attr_name)
    

    def choose_instance_id(self) -> str:
        if hasattr(self, "name"):
            return self.name
        else:
            return str(self.id )

    @classmethod
    def get_logical_key_for_attr(cls, attr_name):
        assert attr_name in cls.get_translateable_attrs()
        return TranslationKeyBuilder.construct_translation_key(
            context=cls.get_translation_context(),
            concept=cls.__name__,
            instance_id=TranslationKeyBuilder.NULL_VALUE,
            attr=attr_name)
    
    @classmethod
    def get_logical_key_for_concept(cls):
        return TranslationKeyBuilder.construct_translation_key(
            context=cls.get_translation_context(),
            concept=cls.__name__,
            instance_id=TranslationKeyBuilder.NULL_VALUE,
            attr=TranslationKeyBuilder.NULL_VALUE)
    

    @classmethod
    @abstractmethod
    def get_translation_context(cls) -> str:
        pass


    @classmethod
    @abstractmethod
    def get_translateable_attrs(cls) -> List[str]:
        pass
